"""Shared metaprogramming helpers for GECCO.
"""

import abc
import bz2
import contextlib
import functools
import importlib
import locale
import lzma
import operator
import typing
from multiprocessing.pool import Pool
from typing import (
    Any,
    BinaryIO,
    Callable,
    Iterable,
    Iterator,
    List,
    Tuple,
    Union,
    Optional,
    Type
)

try:
    from isal import igzip as gzip
except ImportError:
    import gzip

try:
    import lz4.frame
except ImportError as err:
    lz4 = err


if typing.TYPE_CHECKING:
    from types import TracebackType, ModuleType

    _S = typing.TypeVar("_S")
    _T = typing.TypeVar("_T")
    _A = typing.TypeVar("_A")
    _R = typing.TypeVar("_R")
    _F = typing.TypeVar("_F", bound=Callable[..., Any])

_BZ2_MAGIC = b"BZh"
_GZIP_MAGIC = b"\x1f\x8b"
_XZ_MAGIC = b"\xfd7zXZ"
_LZ4_MAGIC = b"\x04\x22\x4d\x18"


class classproperty(property):
    """A class property decorator.

    Example:
        Use `classproperty` to create a property which counts the number of
        times it has been accessed:

        >>> class X():
        ...     __COUNT = 0
        ...     @classproperty
        ...     def count(cls):
        ...         cls.__COUNT += 1
        ...         return cls.__COUNT
        >>> X.count
        1
        >>> X.count
        2

    """

    def __init__(self, f: Callable[["_S"], "_T"]) -> None:
        self.f = f

    def __get__(self, obj: object, owner: "_S") -> "_T":  # type: ignore
        return self.f(owner)


class requires:
    """A decorator for functions that require optional dependencies.
    """

    module: Union["ModuleType", BaseException]

    def __init__(self, module_name: str) -> None:
        self.module_name = module_name

        try:
            self.module = importlib.import_module(module_name)
        except ImportError as err:
            self.module = err

    def __call__(self, func: "_F") -> "_F":

        if isinstance(self.module, ImportError):

            @functools.wraps(func)
            def newfunc(*args, **kwargs):  # type: ignore
                msg = f"calling {func.__qualname__} requires module {self.module.name}"
                raise RuntimeError(msg) from self.module
        else:

            newfunc = func
            basename = self.module_name.split(".")[-1]
            newfunc.__globals__[basename] = self.module

        return newfunc # type: ignore


class UniversalContainer(object):
    """A container that contains everything.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __contains__(self, item: object) -> bool:
        return True


def sliding_window(length: int, window: int, step: int) -> Iterator[slice]:
    """Iterate over a sequence of length `length` with a sliding window.
    """
    if window <= 0:
        raise ValueError("Window size must be strictly positive")
    if step <= 0 or step > window:
        raise ValueError("Window step must be strictly positive and under `window_size`")
    for i in range(0, length + 1 - window, step):
        yield slice(i, i+window)


@contextlib.contextmanager
def patch_locale(name: str) -> Iterator[None]:
    """Create a context manager to locally change the locale in use.
    """
    lc = locale.setlocale(locale.LC_TIME)
    try:
        locale.setlocale(locale.LC_TIME, name)
        yield
    finally:
        locale.setlocale(locale.LC_TIME, lc)


@contextlib.contextmanager
def zopen(path: str) -> Iterator[BinaryIO]:
    with contextlib.ExitStack() as ctx:
        file = ctx.enter_context(open(path, "rb"))
        peek = file.peek()
        if peek.startswith(_GZIP_MAGIC):
            file = ctx.enter_context(gzip.open(file, mode="rb"))
        elif peek.startswith(_BZ2_MAGIC):
            file = ctx.enter_context(bz2.open(file, mode="rb"))
        elif peek.startswith(_XZ_MAGIC):
            file = ctx.enter_context(lzma.open(file, mode="rb"))
        elif peek.startswith(_LZ4_MAGIC):
            if isinstance(lz4, ImportError):
                raise RuntimeError("File compression is LZ4 but python-lz4 is not installed") from lz4
            file = ctx.enter_context(lz4.frame.open(file))
        yield file


import abc
import csv
import errno
import io
import itertools
import math
import operator
import os
import subprocess
import typing
from collections.abc import Sized
from subprocess import DEVNULL
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Type,
    TextIO,
    Union,
    Sequence,
)

import polars

try:
    _POLARS_VERSION = tuple(map(int, polars.__version__.split('.')))
except ValueError:
    _POLARS_VERSION = (0, 0, 0)

if typing.TYPE_CHECKING:
    _SELF = typing.TypeVar("_SELF")
    _TABLE = typing.TypeVar("_TABLE", bound="Table")


class Dumpable(metaclass=abc.ABCMeta):
    """A metaclass for objects that can be dumped to a file.
    """

    @abc.abstractmethod
    def dump(self, fh: Union[BinaryIO, str, os.PathLike]) -> None:
        raise NotImplementedError

    def dumps(self) -> bytes:
        s = io.BytesIO()
        self.dump(s)
        return s.getvalue()


class Loadable(metaclass=abc.ABCMeta):
    """A metaclass for objects that can be loaded from a file.
    """

    @classmethod
    @abc.abstractmethod
    def load(cls: typing.Type["_SELF"], fh: Union[BinaryIO, str, os.PathLike]) -> "_SELF":
        raise NotImplementedError

    @classmethod
    def loads(cls: typing.Type["_SELF"], s: bytes) -> "_SELF":
        return cls.load(io.BytesIO(s))  # type: ignore


class Table(Dumpable, Loadable):
    """A metaclass for objects that
    """

    class Column(typing.NamedTuple):
        name: str
        dtype: type
        default: Optional[object] = None

    @classmethod
    @abc.abstractmethod
    def _get_columns(cls) -> List["Table.Column"]:
        return []
    
    data: polars.DataFrame

    def __init__(self, data: Optional[polars.DataFrame] = None) -> None:
        columns = self._get_columns()

        if data is not None:
            for column in columns:
                if column.name not in data.columns:
                    data = data.with_columns(polars.lit(column.default).alias(column.name))
            self.data = data
        else:
            self.data = polars.DataFrame([
                polars.Series(column.name, dtype=column.dtype)
                for column in columns
            ])

    def __bool__(self) -> bool: # noqa: D105
        return len(self) != 0

    def __len__(self) -> int:
        return len(self.data)
    
    def __getattr__(self, name: str) -> object:
        try:
            return self.data[name]
        except polars.exceptions.ColumnNotFoundError as err:
            raise AttributeError(name) from err

    def __iadd__(self: "_TABLE", rhs: object) -> "_TABLE":  # noqa: D105
        if not isinstance(rhs, type(self)):
            return NotImplemented
        self.data = polars.concat([self.data, rhs.data])
        return self

    @classmethod
    def load(
        cls: typing.Type["_TABLE"], 
        fh: Union[BinaryIO, str, os.PathLike], 
    ) -> "_TABLE":
        columns = cls._get_columns()
        dtypes = { column.name: column.dtype for column in columns }
        if _POLARS_VERSION < (0, 16, 14):
            data = polars.read_csv(fh, sep="\t", dtypes=dtypes)
        else:
            data = polars.read_csv(fh, separator="\t", dtypes=dtypes)
        for column_name in data.columns:
            if data[column_name].dtype in (polars.Float32, polars.Float64):
                data = data.with_columns(polars.col(column_name).fill_null(math.nan))
        return cls(data)

    def dump(self, fh: Union[BinaryIO, str, os.PathLike]) -> None:
        # remove columns that contain only default values
        columns = [ column for column in self._get_columns() ]
        for column in columns.copy():
            if column.default is None:
                continue
            if math.isnan(column.default):
                if self.data[column.name].is_nan().all():
                    columns.remove(column)
            elif self.data[column.name].eq(column.default).all():
                columns.remove(column)
        # write the table as a TSV file
        view = self.data[[column.name for column in columns]]
        for column_name in view.columns:
            if view[column_name].dtype in (polars.Float32, polars.Float64):
                view = view.with_columns(polars.col(column_name).fill_nan(None))
        if _POLARS_VERSION < (0, 16, 14):
            view.write_csv(fh, sep="\t")
        else:
            view.write_csv(fh, separator="\t")
    


"""Generic protocol for ORF detection in DNA sequences.
"""

import abc
import io
import itertools
import os
import queue
import tempfile
import typing
from multiprocessing.pool import Pool, ThreadPool
from multiprocessing.sharedctypes import Value
from typing import Callable, Iterable, Iterator, List, Optional, Tuple, Type, Union

import Bio.SeqIO
import pyrodigal
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .model import Gene, Protein, Strand


__all__ = ["ORFFinder", "PyrodigalFinder"]


class ORFFinder(metaclass=abc.ABCMeta):
    """An abstract base class to provide a generic ORF finder.
    """

    @abc.abstractmethod
    def find_genes(  # type: ignore
        self,
        records: Iterable[SeqRecord],
        progress: Optional[Callable[[SeqRecord, int], None]] = None,
    ) -> Iterable[Gene]:
        """Find all genes from a DNA sequence.
        """
        return NotImplemented


class PyrodigalFinder(ORFFinder):
    """An `ORFFinder` that uses the Pyrodigal bindings to Prodigal.

    Prodigal is a fast and reliable protein-coding gene prediction
    for prokaryotic genomes, with support for draft genomes and metagenomes.

    See Also:
        `Doug Hyatt, Gwo-Liang Chen, Philip F. LoCascio, Miriam L. Land,
        Frank W. Larimer and Loren J. Hauser.
        "Prodigal: Prokaryotic Gene Recognition and Translation Initiation
        Site Identification", BMC Bioinformatics 11 (8 March 2010), p119
        <https://doi.org/10.1186/1471-2105-11-119>`_

    """

    def __init__(self, metagenome: bool = True, mask: bool = False, cpus: int = 0) -> None:
        """Create a new `PyrodigalFinder` instance.

        Arguments:
            metagenome (bool): Whether or not to run Prodigal in metagenome
                mode, defaults to `True`.
            mask (bool): Whether or not to mask genes running across regions
                containing unknown nucleotides, defaults to `False`.
            cpus (int): The number of threads to use to run Pyrodigal in
                parallel. Pass ``0`` to use the number of CPUs on the machine.

        """
        super().__init__()
        self.metagenome = metagenome
        self.mask = mask
        self.cpus = cpus
        self.orf_finder =  pyrodigal.GeneFinder(meta=metagenome, mask=mask)

    def _train(self, records: Iterable[SeqRecord]) -> pyrodigal.TrainingInfo:
        sequences = []
        for i, record in enumerate(records):
            if i > 0:
                sequences.append("TTAATTAATTAA")
            sequences.append(str(record.seq))
        if len(sequences) > 1:
            sequences.append("TTAATTAATTAA")
        return self.orf_finder.train("".join(sequences))

    def _process_record(self, record: SeqRecord) -> Tuple[SeqRecord, pyrodigal.Genes]:
        return record, self.orf_finder.find_genes(str(record.seq))

    def find_genes(
        self,
        records: Iterable[SeqRecord],
        progress: Optional[Callable[[SeqRecord, int], None]] = None,
        *,
        pool_factory: Union[Type[Pool], Callable[[Optional[int]], Pool]] = ThreadPool,
    ) -> Iterator[Gene]:
        """Find all genes contained in a sequence of DNA records.

        Arguments:
            records (iterable of `~Bio.SeqRecord.SeqRecord`): An iterable of
                DNA records in which to find genes.
            progress (callable, optional): A progress callback of signature
                ``progress(record, total)`` that will be called everytime a
                record has been processed successfully, with ``record`` being
                the `~Bio.SeqRecord.SeqRecord` instance, and ``total`` being
                the total number of records to process.

        Keyword Arguments:
            pool_factory (`type`): The callable for creating pools, defaults
                to the `multiprocessing.pool.ThreadPool` class, but
                `multiprocessing.pool.Pool` is also supported.

        Yields:
            `~gecco.model.Gene`: An iterator over all the genes found in
            the given records.

        """
        # detect the number of CPUs
        _cpus = self.cpus if self.cpus > 0 else os.cpu_count()
        _progress = (lambda x,y: None) if progress is None else progress

        # train first if needed
        if not self.metagenome:
            records = list(records)
            self._train(records)

        # run in parallel using a pool
        with pool_factory(_cpus) as pool:
            for record, orfs in pool.imap(self._process_record, records):
                _progress(record, len(orfs))
                for j, orf in enumerate(orfs):
                    # wrap the protein into a Protein object
                    prot_seq = Seq(orf.translate())
                    protein = Protein(id=f"{record.id}_{j+1}", seq=prot_seq)
                    # wrap the gene into a Gene
                    yield Gene(
                        source=record,
                        start=min(orf.begin, orf.end),
                        end=max(orf.begin, orf.end),
                        strand=Strand(orf.strand),
                        protein=protein,
                        qualifiers={
                            "inference": [f"ab initio prediction:Pyrodigal:{pyrodigal.__version__}"],
                            "transl_table": [str(orf.translation_table)],
                        }
                    )


class CDSFinder(ORFFinder):
    """An `ORFFinder` that simply extracts CDS annotations from records.
    """

    def __init__(
        self,
        feature: str = "CDS",
        translation_table: int = 11,
        locus_tag: str = "locus_tag",
    ):
        self.feature = feature
        self.translation_table = translation_table
        self.locus_tag = locus_tag

    def find_genes(
        self,
        records: Iterable[SeqRecord],
        progress: Optional[Callable[[SeqRecord, int], None]] = None,
    ) -> Iterator[Gene]:
        """Find all genes contained in a sequence of DNA records.
        """
        ids = set()
        _progress = (lambda x,y: None) if progress is None else progress

        for record in records:
            genes_found = 0
            features = filter(lambda feat: feat.type == self.feature, record.features)
            for i, feature in enumerate(features):
                # get the gene translation
                tt = feature.qualifiers.get("transl_table", [self.translation_table])[0]
                if "translation" in feature.qualifiers:
                    prot_seq = Seq(feature.qualifiers["translation"][0])
                else:
                    prot_seq = feature.location.extract(record.seq).translate(table=tt)
                # get the gene name
                if self.locus_tag in feature.qualifiers:
                    protein = Protein(id=feature.qualifiers[self.locus_tag][0], seq=prot_seq)
                else:
                    protein = Protein(id=f"{record.id}_{i+1}", seq=prot_seq)
                # check IDs are unique
                if protein.id in ids:
                    raise ValueError(f"Duplicate gene identifier found in {record.id!r}: {protein.id!r}")
                ids.add(protein.id)
                # wrap the gene into a Gene
                yield Gene(
                    source=record,
                    start=feature.location.start + 1,
                    end=feature.location.end,
                    strand=Strand(feature.location.strand),
                    protein=protein,
                )
                genes_found += 1
            _progress(record, genes_found)



"""Gene cluster prediction using a conditional random field.
"""

import csv
import copy
import functools
import hashlib
import itertools
import io
import math
import numbers
import operator
import os
import pickle
import random
import textwrap
import typing
import warnings
from typing import (
    Any,
    BinaryIO,
    Callable,
    ContextManager,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    Union,
)

import numpy
import tqdm
import sklearn_crfsuite
import sklearn.model_selection
import sklearn.preprocessing

from .._meta import sliding_window
from ..model import Gene
from . import features
from .cv import LeaveOneGroupOut
from .select import fisher_significance

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files  # type: ignore

__all__ = ["ClusterCRF"]


class ClusterCRF(object):
    """A wrapper for `sklearn_crfsuite.CRF` to work with the GECCO data model.
    """

    _FILENAME = "model.pkl"

    @classmethod
    def trained(cls, model_path: Optional[str] = None) -> "ClusterCRF":
        """Create a new pre-trained `ClusterCRF` instance from a model path.

        Arguments:
            model_path (`str`, optional): The path to the model directory
                obtained with the ``gecco train`` command. If `None` given,
                use the embedded model.

        Returns:
            `~gecco.crf.ClusterCRF`: A CRF model that can be used to perform
            predictions without training first.

        Raises:
            `ValueError`: If the model data does not match its hash.

        """
        # get the path to the pickled model and read its signature file
        if model_path is not None:
            pkl_file: ContextManager[BinaryIO] = open(os.path.join(model_path, cls._FILENAME), "rb")
            md5_file: ContextManager[TextIO] = open(os.path.join(model_path, f"{cls._FILENAME}.md5"))
        else:
            pkl_file = files(__name__).joinpath(cls._FILENAME).open("rb")
            md5_file = files(__name__).joinpath(f"{cls._FILENAME}.md5").open()
        with md5_file as sig:
            signature = sig.read().strip()

        # check the file content matches its MD5 hashsum
        hasher = hashlib.md5()
        with pkl_file as bin:
            read = functools.partial(bin.read, io.DEFAULT_BUFFER_SIZE)
            for chunk in iter(read, b""):
                hasher.update(typing.cast(bytes, chunk))
            if hasher.hexdigest().upper() != signature.upper():
                raise ValueError("MD5 hash of model data does not match signature")
            pkl_file.seek(0)  # type: ignore
            return pickle.load(bin)  # type: ignore

    def __init__(
        self,
        feature_type: str = "protein",
        algorithm: str = "lbfgs",
        window_size: int = 5,
        window_step: int = 1,
        **kwargs: Any,
    ) -> None:
        """Create a new `ClusterCRF` instance.

        Arguments:
            feature_type (`str`): Defines how features should be extracted.
                Should be either *domain* or *protein*.
            algorithm (`str`): The optimization algorithm for the model. See
                https://sklearn-crfsuite.readthedocs.io/en/latest/api.html
                for available values.
            window_size (`int`): The size of the sliding window to use
                when training and predicting probabilities on sequences
                of genes.
            window_step (`int`): The step between consecutive sliding
                windows to use when training and predicting probabilities
                on sequences of genes.

        Any additional keyword argument is passed as-is to the internal
        `~sklearn_crfsuite.CRF` constructor.

        Raises:
            `ValueError`: if ``feature_type`` has an invalid value.
            `TypeError`: if one of the ``*_columns`` argument is not iterable.

        """
        if feature_type not in {"protein", "domain"}:
            raise ValueError(f"invalid feature type: {feature_type!r}")
        if window_size <= 0:
            raise ValueError("Window size must be strictly positive")
        if window_step <= 0 or window_step > window_size:
            raise ValueError("Window step must be strictly positive and under `window_size`")

        self.feature_type = feature_type
        self.window_size = window_size
        self.window_step = window_step
        self.algorithm = algorithm
        self.significance: Optional[Dict[str, float]] = None
        self.significant_features: Optional[FrozenSet[str]] = None
        self.model = None
        self._options = {"algorithm": algorithm, **kwargs}

    def predict_probabilities(
        self,
        genes: Iterable[Gene],
        *,
        pad: bool = True,
        progress: Optional[Callable[[int, int], None]] = None,
    ) -> List[Gene]:
        """Predict how likely each given gene is part of a gene cluster.

        Arguments:
            genes (iterable of `~gecco.model.Gene`): The genes to compute
                probabilities for.

        Keyword Arguments:
            batch_size (`int`): The number of samples to load per batch.
                *Ignored, always 1 with the CRF.*
            pad (`bool`): Whether to pad sequences too small for a single
                window. Setting this to `False` will skip probability
                prediction entirely for sequences smaller than the window
                size.
            progress (callable): A callable that accepts two `int`, the
                current batch index and the total number of batches.

        Returns:
            `list` of `~gecco.model.Gene`: A list of new `Gene` objects with
            their probability set.

        Raises:
            `~sklearn.exceptions.NotFittedError`: When calling this method
                on an object that has not been fitted yet.

        """
        # silence progress if no callback given, ignored
        _progress = progress or (lambda x,y: None)
        window_index = 0

        # check that the model was trained
        if self.model is None:
            raise NotFittedError("This ClusterCRF instance is not fitted yet.")

        # select the feature extraction method
        if self.feature_type == "protein":
            extract_features = features.extract_features_protein
            annotate_probabilities = features.annotate_probabilities_protein
        elif self.feature_type == "domain":
            extract_features = features.extract_features_domain
            annotate_probabilities = features.annotate_probabilities_domain
        else:
            raise ValueError(f"invalid feature type: {self.feature_type!r}")

        # sort genes by sequence id and domains inside genes by coordinate
        genes = sorted(genes, key=operator.attrgetter("source.id", "start"))
        for gene in genes:
            gene.protein.domains.sort(key=operator.attrgetter("start"))

        # group genes by contigs
        contigs = {}
        for contig_id, group in itertools.groupby(genes, key=operator.attrgetter("source.id")):
            contigs[contig_id] = list(group)

        # extract features from all contigs
        contig_features = {}
        deltas = {}
        for contig_id, contig in contigs.items():
            # extract features
            feats: List[Dict[str, bool]] = extract_features(contig)
            deltas[contig_id] = 0
            # ignore sequences too small with a warning
            if len(feats) < self.window_size:
                if pad:
                    unit = self.feature_type if self.window_size - len(feats) == 1 else f"{self.feature_type}s"
                    warnings.warn(
                        f"Contig {contig[0].source.id!r} does not contain enough"
                        f" {self.feature_type}s ({len(contig)}) for sliding window"
                        f" of size {self.window_size}, padding with"
                        f" {self.window_size - len(feats)} {unit}"
                    )
                    # insert empty features as padding on both ends
                    deltas[contig_id] = delta = self.window_size - len(feats)
                    feats = [{} for _ in range(delta // 2)] + feats + [{} for _ in range((delta+1)//2)]
                else:
                    warnings.warn(
                        f"Contig {contig[0].source.id!r} does not contain enough"
                        f" {self.feature_type}s ({len(contig)}) for sliding window"
                        f" of size {self.window_size}"
                    )
                    continue
            # store features for the current contig
            contig_features[contig_id] = feats

        # compute total number of windows to process
        total = sum(len(feats) - self.window_size + 1 for feats in contig_features.values())
        _progress(window_index, total)

        # predict probabilities
        predicted = []
        for contig_id, contig in contigs.items():
            # get features if the sequence was not skipped
            if contig_id not in contig_features:
                predicted.extend(contig)
                continue
            feats = contig_features[contig_id]
            # predict marginals over a sliding window, storing maximum probabilities
            probabilities = numpy.zeros(max(len(contig), self.window_size))
            for win in sliding_window(len(feats), self.window_size, self.window_step):
                marginals = [p['1'] for p in self.model.predict_marginals_single(feats[win])]
                numpy.maximum(probabilities[win], marginals, out=probabilities[win])
                window_index += 1
                _progress(window_index, total)
            # label genes with maximal probabilities
            predicted.extend(annotate_probabilities(contig, probabilities[deltas[contig_id]//2:][:len(contig)]))

        # label domains with their biosynthetic weight according to the CRF state weights
        predicted = [
            gene.with_protein(gene.protein.with_domains(
                domain.with_cluster_weight(
                    self.model.state_features_.get((domain.name, '1'))
                )
                for domain in gene.protein.domains
            ))
            for gene in predicted
        ]

        # return the genes that were passed as input but now having 
        # gene cluster probabilities
        return predicted

    def fit(
        self,
        genes: Iterable[Gene],
        *,
        select: Optional[float] = None,
        shuffle: bool = True,
        cpus: Optional[int] = None,
        correction_method: Optional[str] = None,
    ) -> None:
        """Fit the CRF model to the given training data.

        Arguments:
            genes (iterable of `~gecco.model.Gene`): The genes to extract
                domains from for training the CRF.
            select (`float`, *optional*): The fraction of features to
                select based on Fisher-tested significance. Leave as `None`
                to skip feature selection.
            shuffle (`bool`): Whether or not to shuffle the contigs
                after having grouped the genes together.
            correction_method (`str`, *optional*): The correction method
                to use for correcting p-values used for feature selection.
                Ignored if ``select`` is `False`.

        """
        _cpus = os.cpu_count() if not cpus else cpus
        # select the feature extraction method
        if self.feature_type == "protein":
            extract_features = features.extract_features_protein
            extract_labels = features.extract_labels_protein
        elif self.feature_type == "domain":
            extract_features = features.extract_features_domain
            extract_labels = features.extract_labels_domain
        else:
            raise ValueError(f"invalid feature type: {self.feature_type!r}")

        # sort genes by sequence id and domains inside genes by coordinate
        genes = sorted(genes, key=operator.attrgetter("source.id"))
        for gene in genes:
            gene.protein.domains.sort(key=operator.attrgetter("start"))

        # perform feature selection
        if select is not None:
            if select <= 0 or select > 1:
                raise ValueError(f"invalid value for select: {select}")
            # find most significant features
            self.significance = sig = fisher_significance(
                (gene.protein for gene in genes),
                correction_method=correction_method,
            )
            sorted_sig = sorted(sig, key=sig.get)[:int(select*len(sig))]  # type: ignore
            self.significant_features = frozenset(sorted_sig)
            # check that we don't select "random" domains
            if sig[sorted_sig[-1]] == 1.0:
                warnings.warn(
                    "Selected features still include domains with a p-value "
                    "of 1, consider reducing the selected fraction.",
                    UserWarning
                )
            # remove non significant domains
            genes = [
                gene.with_protein(
                    gene.protein.with_domains([
                        domain for domain in gene.protein.domains
                        if domain.name in self.significant_features
                    ])
                )
                for gene in genes
            ]

        # group input genes by sequence
        groups = itertools.groupby(genes, key=operator.attrgetter("source.id"))
        sequences = [sorted(group, key=operator.attrgetter("start")) for _, group in groups]
        # shuffle sequences if requested
        if shuffle:
            random.shuffle(sequences)

        # build training instances
        training_features, training_labels = [], []
        for sequence in sequences:
            # extract features and labels
            feats: List[Dict[str, bool]] = extract_features(sequence)
            labels: List[str] = extract_labels(sequence)
            # check we have as many observations as we have labels
            if len(feats) != len(labels):
                raise ValueError("different number of features and labels found, something is wrong")
            # check we have enough genes for the desired sliding window
            if len(feats) < self.window_size:
                raise ValueError(f"{sequence[0].source.id!r} has not enough observations ({len(feats)}) for requested window size ({self.window_size})")
            # record every window in the sequence
            for win in sliding_window(len(feats), self.window_size, self.window_step):
                training_features.append(feats[win])
                training_labels.append(labels[win])

        # check labels
        if all(label == "1" for y in training_labels for label in y):
            raise ValueError("only positives labels found, something is wrong.")
        elif all(label == "0" for y in training_labels for label in y):
            raise ValueError("only negative labels found, something is wrong.")

        # fit the model
        self.model = model = sklearn_crfsuite.CRF(**self._options)
        model.fit(training_features, training_labels)

    def save(self, model_path: "os.PathLike[str]") -> None:
        """Save the `ClusterCRF` to an on-disk location.

        Models serialized at a given location can be later loaded from that
        same location using the `ClusterCRF.trained` class method.

        Arguments:
            model_path (`str`): The path to the directory where to write
                the model files.

        """
        # pickle the CRF model
        model_out = os.path.join(model_path, self._FILENAME)
        with open(model_out, "wb") as out:
            pickle.dump(self, out, protocol=4)
        # compute a checksum for the pickled model
        hasher = hashlib.md5()
        with open(model_out, "rb") as out:
            for chunk in iter(lambda: out.read(io.DEFAULT_BUFFER_SIZE), b""):
                hasher.update(chunk)
        # write the checksum next to the pickled file
        with open(f"{model_out}.md5", "w") as out_hash:
            out_hash.write(hasher.hexdigest())



# coding: utf-8
import abc
import contextlib
import datetime
import logging
import os
import socket
import sys
import textwrap
import typing
from typing import Any, ClassVar, Callable, Optional, List, Mapping, Dict, TextIO, Type

import docopt
import rich.console
import rich.progress
import rich.logging
from rich.text import Text

from ... import __version__, __name__ as __progname__

_T = typing.TypeVar("_T")


class InvalidArgument(ValueError):
    """An error to mark an invalid value was passed to a CLI flag.
    """

class CommandExit(Exception):
    """An error to request immediate exit from a function.
    """

    def __init__(self, code: int) -> None:
        self.code = code


class MofNWithPrecisionColumn(rich.progress.MofNCompleteColumn):

    def render(self, task: "rich.progress.Task") -> "Text":
        """Show completed/total."""
        completed = task.completed
        total = task.total if task.total is not None else "?"
        precision = task.fields.get("precision", "")
        total_width = len(f"{total:{precision}}")
        return Text(
            f"{completed:{total_width}{precision}}{self.separator}{total:{precision}}",
            style="progress.download",
        )


class Command(metaclass=abc.ABCMeta):
    """An abstract base class for ``gecco`` subcommands.
    """

    # -- Abstract methods ----------------------------------------------------

    summary: ClassVar[str] = NotImplemented

    @abc.abstractmethod
    def execute(self, ctx: contextlib.ExitStack) -> int:
        """Execute the command.

        Returns:
            `int`: The exit code for the command, with 0 on success, and any
            other number on error.

        """
        return NotImplemented

    @classmethod
    @abc.abstractmethod
    def doc(cls, fast: bool = False) -> str:
        """Get the help message for the command.
        """
        return NotImplemented

    # -- Concrete methods ----------------------------------------------------

    _version: str = "{} {}".format(__progname__, __version__)
    _options_first: bool = False

    def __init__(
        self,
        argv: Optional[List[str]] = None,
        stream: Optional[TextIO] = None,
        options: Optional[Mapping[str, Any]] = None,
        config: Optional[Dict[Any, Any]] = None,
    ) -> None:
        self._stream: Optional[TextIO] = stream
        self.argv = argv
        self.stream: TextIO = stream or sys.stderr
        self.options = options or dict()
        self.pool = None
        self.config = config

        self._hostname = socket.gethostname()
        self._pid = os.getpid()

        # Parse command line arguments
        try:
            self.args = docopt.docopt(
                textwrap.dedent(self.doc(fast=True)).lstrip(),
                help=False,
                argv=argv,
                version=self._version,
                options_first=self._options_first,
            )
            self.verbose = self.args.get("--verbose", 0)
            self.quiet = self.args.get("--quiet", 0)
        except docopt.DocoptExit as de:
            self.args = de
            self.level = 0
            self.quiet = 0

        self.progress = rich.progress.Progress(
            rich.progress.SpinnerColumn(finished_text="[green]:heavy_check_mark:[/]"),
            "[progress.description]{task.description}",
            rich.progress.BarColumn(bar_width=60),
            MofNWithPrecisionColumn(),
            "[progress.completed]{task.fields[unit]}",
            "[progress.percentage]{task.percentage:>3.0f}%",
            rich.progress.TimeElapsedColumn(),
            rich.progress.TimeRemainingColumn(),
            console=rich.console.Console(file=self.stream, soft_wrap=True),
            disable=self.quiet > 0,
            transient=True,
        )
        self.console = self.progress.console

    def _check(self) -> None:
        # Assert CLI arguments were parsed Successfully
        if isinstance(self.args, docopt.DocoptExit):
            self.console.print(self.args)#, file=self.stream)
            raise CommandExit(1)
        return None

    def _check_flag(
        self,
        name: str,
        convert: Optional[Callable[[str], _T]] = None,
        check: Optional[Callable[[_T], bool]] = None,
        message: Optional[str] = None,
        hint: Optional[str] = None,
        optional: bool = False,
        default: _T = None
    ) -> _T:
        _convert = (lambda x: x) if convert is None else convert
        _check = (lambda x: True) if check is None else check
        if self.args[name] is None:
            if optional:
                return default  # type: ignore
            self.error(f"Missing value for argument [purple]{name}[/]")
            raise InvalidArgument(self.args[name])
        try:
            value = _convert(self.args[name])
            if not _check(value):
                raise ValueError(self.args[name])
        except Exception as err:
            if hint is None:
                self.error(f"Invalid value for argument [purple]{name}[/]:", repr(self.args[name]))
            else:
                self.error(f"Invalid value for argument [purple]{name}[/]:", repr(self.args[name]), f"(expected {hint})")
            raise InvalidArgument(self.args[name]) from err
        else:
            return value  # type: ignore

    def _on_import_error(
        self,
        subcommand: str,
        e: ImportError
    ) -> None:
        import rich.traceback

        self.error(f"The [bold blue]{subcommand}[/] subcommand requires optional dependency [bold blue]{e.name}[/]")
        traceback = rich.traceback.Traceback.from_exception(type(e), e, e.__traceback__, extra_lines=0)
        self.console.print(traceback)


    # -- Logging methods -----------------------------------------------------

    def error(self, message: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 2 and level <= self.verbose:
            if self.verbose <= 1:
                self.console.print(
                    "[bold red]x[/]",
                    message,
                    *args,
                )
            else:
                self.console.print(
                    *self._logprefix(),
                    "[bold red]FAIL[/]",
                    message,
                    *args,
                )

    def info(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            if self.verbose <= 1:
                self.console.print(
                    "[bold blue]i[/]",
                    verb,
                    *args,
                )
            else:
                self.console.print(
                    *self._logprefix(),
                    f"[bold blue]INFO[/]",
                    verb,
                    *args,
                )

    def success(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            if self.verbose <= 1:
                self.console.print(
                    "[green]:heavy_check_mark:[/]",
                    verb,
                    *args,
                )
            else:
                self.console.print(
                    *self._logprefix(),
                    f"[bold green]  OK[/]",
                    verb,
                    *args,
                )

    def warn(self, verb: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 1 and level <= self.verbose:
            if self.verbose <= 1:
                self.console.print(
                    "[bold yellow]![/]",
                    verb,
                    *args,
                )
            else:
                self.console.print(
                    *self._logprefix(),
                    "[bold yellow]WARN[/]",
                    verb,
                    *args
                )

    def _logprefix(self) -> List[str]:
        return [
            f"[dim cyan]{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}[/]",
            f"[dim purple]{self._hostname}[/]",
            f"[dim]{__progname__}[[default dim]{self._pid}[/]][/]",
        ]

    def _showwarnings(
        self,
        message: str,
        category: Type[Warning],
        filename: str,
        lineno: int,
        file: Optional[TextIO] = None,
        line: Optional[str] = None,
    ) -> None:
        for line in filter(str.strip, str(message).splitlines()):
            self.warn(line.strip())



"""Implementation of the ``gecco rerun`` subcommand.
"""

import contextlib
import errno
import glob
import io
import itertools
import json
import logging
import multiprocessing
import operator
import os
import pickle
import tempfile
import typing
import signal
from typing import Any, Dict, Union, Optional, Iterable, List, TextIO, Mapping

import rich.emoji
import rich.progress

from ... import __version__
from ._base import Command, CommandExit, InvalidArgument
from ._mixins import TableLoaderMixin, SequenceLoaderMixin, OutputWriterMixin, DomainFilterMixin, PredictorMixin
from .._utils import patch_showwarnings

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources  # type: ignore

if typing.TYPE_CHECKING:
    from ...types import TypeClassifier
    from ...model import Cluster, Gene, FeatureTable, GeneTable
    from Bio.SeqIO import SeqRecord


class Predict(TableLoaderMixin, SequenceLoaderMixin, OutputWriterMixin, DomainFilterMixin, PredictorMixin):  # noqa: D101

    summary = "predict gene clusters on contigs that have already been annotated."

    @classmethod
    def doc(cls, fast: bool = False) -> str:  # noqa: D102
        return f"""
        gecco predict - {cls.summary}

        Usage:
            gecco predict --genome <file> --features <table>... --genes <table> [options]

        Arguments:
            -g <file>, --genome <file>    a genomic file containing one or more
                                          sequences to use as input. Must be in
                                          one of the sequences format supported
                                          by Biopython.
            --features <table>            a feature table obtained by a previous
                                          invocation of ``gecco run``.
            --genes <table>               a gene table obtained by a previous
                                          invocation of ``gecco run``.


        Parameters:
            -f <fmt>, --format <fmt>      the format of the input file, as a
                                          Biopython format string. GECCO is able
                                          to recognize FASTA and GenBank files
                                          automatically if this is not given.
            -j <jobs>, --jobs <jobs>      the number of CPUs to use for
                                          multithreading. Use 0 to use all of the
                                          available CPUs. [default: 0]

        Parameters - Output:
            -o <out>, --output-dir <out>  the directory in which to write the
                                          output files. [default: .]
            --antismash-sideload          write an AntiSMASH v6 sideload JSON
                                          file next to the output files.
            --force-tsv                   always write TSV output files even
                                          when they are empty (e.g. because
                                          no genes or no clusters were found).
            --merge-gbk                   output a single file containing
                                          every detected cluster instead of
                                          writing one file per cluster.

        Parameters - Domain Annotation:
            -e <e>, --e-filter <e>        the e-value cutoff for protein domains
                                          to be included. This is not stable
                                          across versions, so consider using
                                          a p-value filter instead.
            -p <p>, --p-filter <p>        the p-value cutoff for protein domains
                                          to be included. [default: 1e-9]

        Parameters - Cluster Detection:
            --no-pad                      disable padding of gene sequences
                                          (used to predict gene clusters in 
                                          contigs smaller than the CRF window 
                                          length).
            -c <N>, --cds <N>             the minimum number of coding sequences a
                                          valid cluster must contain. [default: 3]
            -m <m>, --threshold <m>       the probability threshold for cluster
                                          detection. Default depends on the
                                          post-processing method (0.8 for gecco,
                                          0.6 for antismash).
            --postproc <method>           the method to use for cluster validation
                                          (antismash or gecco). [default: gecco]
            -E <N>, --edge-distance <N>   the minimum number of annotated genes
                                          that must separate a cluster from the
                                          edge. Edge clusters will still be
                                          included if they are longer. A lower
                                          number will increase the number of
                                          false positives on small contigs.
                                          [default: 0]

        Parameters - Debug:
            --model <directory>           the path to an alternative CRF model
                                          to use (obtained with `gecco train`).

        """

    def _check(self) -> None:
        Command._check(self)
        try:
            self.cds = self._check_flag("--cds", int, lambda x: x > 0, hint="positive integer")
            self.e_filter = self._check_flag(
                "--e-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.p_filter = self._check_flag(
                "--p-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            if self.args["--threshold"] is None:
                self.threshold = 0.8 if self.args["--postproc"] == "gecco" else 0.6
            else:
                self.threshold = self._check_flag("--threshold", float, lambda x: 0 <= x <= 1, hint="number between 0 and 1")
            self.jobs = self._check_flag("--jobs", int, lambda x: x >= 0, hint="positive or null integer")
            self.postproc = self._check_flag("--postproc", str, lambda x: x in ("gecco", "antismash"), hint="'gecco' or 'antismash'")
            self.edge_distance = self._check_flag("--edge-distance", int, lambda x: x >= 0, hint="positive or null integer")
            self.format = self._check_flag("--format", optional=True)
            self.genome = self._check_flag("--genome")
            self.features: List[str] = self._check_flag("--features")
            self.genes: str = self._check_flag("--genes")
            self.model: Optional[str] = self._check_flag("--model", optional=True)
            self.output_dir: str = self._check_flag("--output-dir")
            self.antismash_sideload = self._check_flag("--antismash-sideload", bool)
            self.force_tsv = self._check_flag("--force-tsv", bool)
            self.no_pad = self._check_flag("--no-pad", bool)
            self.merge_gbk = self._check_flag("--merge-gbk", bool)
        except InvalidArgument:
            raise CommandExit(1)

    # ---

    def _assign_sources(self, sequences: Iterable["SeqRecord"], genes: List["Gene"]) -> Iterable["Gene"]:
        from ...model import Protein, Strand

        self.info("Extracting", "known sequences from gene list", level=2)
        known_sequences = { gene.source.id for gene in genes }

        self.info("Building", "index of source sequences", level=2)
        sequence_index = {
            record.id:record
            for record in sequences
            if record.id in known_sequences
        }

        try:
            self.info("Assigning", "source sequences to gene objects", level=2)
            for gene in genes:
                gene = gene.with_source(sequence_index[gene.source.id])
                gene_seq = gene.source.seq[gene.start-1:gene.end]
                if gene.strand == Strand.Reverse:
                    gene_seq = gene_seq.reverse_complement()
                gene = gene.with_protein(gene.protein.with_seq(gene_seq.translate()))
                yield gene
        except KeyError as err:
            self.error("Sequence {!r} not found in {!r}", gene.source.id, self.genome)
            raise CommandExit(1) from err

    def _write_sideload_json(self, clusters: List["Cluster"]) -> None:
        # record version and important parameters
        data: Dict[str, Any] = {
            "records": [],
            "tool": {
                "name": "GECCO",
                "version": __version__,
                "description": "Biosynthetic Gene Cluster prediction with Conditional Random Fields.",
                "configuration": {
                    "cds": repr(self.cds),
                    "e-filter": None if self.e_filter is None else repr(self.e_filter),
                    "p-filter": None if self.p_filter is None else repr(self.p_filter),
                    "postproc": repr(self.postproc),
                    "threshold": repr(self.threshold),
                    "edge-distance": repr(self.edge_distance),
                    "no-pad": repr(self.no_pad),
                }
            }
        }
        # record if non-standard HMM or model was used
        if self.model:
            data["tool"]["configuration"]["model"] = self.model
        # create a record per sequence
        for seq_id, seq_clusters in itertools.groupby(clusters, key=operator.attrgetter("source.id")):
            data["records"].append({"name": seq_id, "subregions": []})
            for cluster in seq_clusters:
                probabilities = {
                    f"{key.lower()}_probability":f"{value:.3f}"
                    for key, value in cluster.type_probabilities.items()
                }
                data["records"][-1]["subregions"].append({
                    "start": cluster.start,
                    "end": cluster.end,
                    "label": ";".join(sorted(cluster.type.names)) or "Unknown",
                    "details": {
                        "average_p": f"{cluster.average_probability:.3f}",
                        "max_p": f"{cluster.maximum_probability:.3f}",
                        **probabilities,
                    }
                })
        # write the JSON file to the output folder
        base, _ = os.path.splitext(os.path.basename(self.genome))
        sideload_out = os.path.join(self.output_dir, f"{base}.sideload.json")
        self.info("Writing", "sideload JSON to", repr(sideload_out), level=1)
        with open(sideload_out, "w") as out:
            json.dump(data, out, sort_keys=True, indent=4)

    # ---

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check the CLI arguments were fine and enter context
            self._check()
            ctx.enter_context(self.progress)
            ctx.enter_context(patch_showwarnings(self._showwarnings))  # type: ignore
            # attempt to create the output directory, checking it doesn't
            # already contain output files (or raise a warning)
            base, _ = os.path.splitext(os.path.basename(self.genome))
            outputs = [f"{base}.clusters.tsv", f"{base}.features.tsv", f"{base}.genes.tsv"]
            if self.antismash_sideload:
                outputs.append(f"{base}.sideload.json")
            self._make_output_directory(outputs)
            # load features
            genes = list(self._load_genes())
            features = self._load_features()
            # load sequences
            sequences = self._load_sequences()
            # label genes
            genes = self._annotate_genes(genes, features)
            genes = list(self._assign_sources(sequences, genes))
            # Sort genes
            self.info("Sorting", "genes by coordinates", level=2)
            genes.sort(key=operator.attrgetter("source.id", "start", "end"))
            for gene in genes:
                gene.protein.domains.sort(key=operator.attrgetter("start", "end"))
            # filter domains by p-value and/or e-value
            genes = self._filter_domains(genes)
            # check if probabilities need to be predicted
            if not all(gene.average_probability is not None for gene in genes):
                genes = self._predict_probabilities(genes)
            self._write_genes_table(genes)
            self._write_feature_table(genes)
            # extract clusters from probability vector
            clusters = self._extract_clusters(genes)
            if clusters:
                self.success("Found", len(clusters), "potential gene clusters", level=1)
            else:
                self.warn("No gene clusters were found")
                if self.force_tsv:
                    self._write_cluster_table(clusters)
                return 0
            # predict types for putative clusters
            classifier = self._load_type_classifier()
            if len(classifier.classes_) > 1:
                clusters = self._predict_types(clusters, classifier)
            # write results
            self.info("Writing", "result files to folder", repr(self.output_dir), level=1)
            self._write_cluster_table(clusters)
            self._write_clusters(clusters, merge=self.merge_gbk)
            if self.antismash_sideload:
                self._write_sideload_json(clusters)
            unit = "cluster" if len(clusters) == 1 else "clusters"
            self.success("Found", len(clusters), "gene", unit, level=0)
        except CommandExit as cexit:
            return cexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except Exception as err:
            self.progress.stop()
            raise
        else:
            return 0



"""Features extraction and sanitization utilities for `~gecco.crf.ClusterCRF`.
"""
import itertools
import typing
from typing import Iterable, List, Sequence, Dict

from ..model import Gene

if typing.TYPE_CHECKING:
    _S = typing.TypeVar("_S", bound=Sequence[Gene])


def extract_features_protein(
    sequence: Iterable[Gene], empty: bool = True
) -> List[Dict[str, bool]]:
    """Extract features at the gene level.

    If several domains are part of the same gene, they are grouped at the
    same position. Genes without domain annotation are represented with an
    empty dictionary.

    Arguments:
        sequence (iterable of `~gecco.model.Gene`): The genes to extract the
            features from.
        empty (`bool`): Whether or not to skip empty genes. With `False`,
            no empty feature dictionary will be emitted, and genes will be
            skipped. With `True`, an empty feature dictionary will be emitted
            for every gene without domain annotation.

    """
    return [
        {domain.name: True for domain in gene.protein.domains}
        for gene in sequence
        if gene.protein.domains or empty
    ]


def extract_features_domain(
    sequence: Iterable[Gene], empty: bool = True
) -> List[Dict[str, bool]]:
    """Extract features at the domain level."""
    features: List[Dict[str, bool]] = []
    for gene in sequence:
        if gene.protein.domains:
            features.extend({domain.name: True} for domain in gene.protein.domains)
        elif empty:
            features.append({})
    return features


def extract_labels_protein(sequence: Iterable[Gene], empty: bool = True) -> List["str"]:
    """Extract labels at the gene level for training."""
    return [
        "1" if typing.cast(float, gene.average_probability) > 0.5 else "0"
        for gene in sequence
        if gene.protein.domains or empty
    ]


def extract_labels_domain(sequence: Iterable[Gene], empty: bool = True) -> List["str"]:
    """Extract labels at the domain level for training."""
    labels: List[str] = []
    for gene in sequence:
        if gene.protein.domains:
            labels.extend(
                "1" if typing.cast(float, domain.probability) > 0.5 else "0"
                for domain in gene.protein.domains
            )
        elif empty:
            labels.append("1" if typing.cast(float, gene.average_probability) > 0.5 else "0")
    return labels


def annotate_probabilities_protein(
    sequence: Sequence[Gene],
    probabilities: Sequence[float],
    empty: bool = True,
) -> Iterable[Gene]:
    """Annotate genes with marginals obtained from a CRF at the protein level.

    Arguments:
        sequence (iterable of `~gecco.model.Gene`): The genes to annotate the
            features from.
        probabilities (iterable of `float`): The biosynthetic probabilities
            that have been computed by the CRF.
        empty (`bool`): Whether or not to empty genes where skipped. With
            `False`, no empty feature dictionary have been emitted, so empty
            genes have no probability. With `True`, empty genes have a
            probability that can be extracted.

    """
    genes = [gene for gene in sequence if gene.protein.domains or empty]
    if len(genes) != len(probabilities):
        raise ValueError("gene and probability lists don't have the same length")
    for gene, probability in zip(genes, probabilities):
        yield gene.with_probability(probability)


def annotate_probabilities_domain(
    sequence: Iterable[Gene],
    probabilities: Iterable[float],
    empty: bool = True,
) -> Iterable[Gene]:
    """Annotate genes with marginals obtained from a CRF at the domain level."""
    genes = iter(sequence)
    probas = iter(probabilities)
    for gene in genes:
        if gene.protein.domains:
            yield gene.with_protein(
                gene.protein.with_domains(
                    [
                        domain.with_probability(p)
                        for domain, p in zip(gene.protein.domains, probas)
                    ]
                )
            )
        elif empty:
            yield gene.with_probability(next(probas))
    if next(probas, None) is not None:
        raise ValueError("gene and probability lists don't have the same length")



"""Data layer classes storing information needed for gene cluster detection.
"""

import collections
import copy
import csv
import datetime
import enum
import functools
import itertools
import math
import operator
import os
import re
import statistics
import typing
from array import array
from collections import OrderedDict
from collections.abc import Sized
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Mapping, Optional, Sequence, BinaryIO, NamedTuple, Union, Iterator, Set
from typing import Dict, Iterable, List, Mapping, Optional, Sequence, BinaryIO, NamedTuple, Union, Iterator

import Bio
import numpy
import polars
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation, Reference
from Bio.SeqRecord import SeqRecord

from . import __version__
from .interpro import GOTerm
from ._base import Dumpable, Table, _POLARS_VERSION
from ._meta import patch_locale

if typing.TYPE_CHECKING:
    from numpy.typing import NDArray


__all__ = [
    "ClusterType",
    "Strand",
    "Domain",
    "Protein",
    "Gene",
    "Cluster",
    "FeatureTable",
    "ClusterTable"
]


# fmt: off
class ClusterType(object):
    """An immutable storage for the type of a gene cluster.
    """

    def __init__(self, *names: str) -> None:
        """Create a new product type from one or more base types.

        Example:
            >>> t1 = ClusterType()                    # unknown type
            >>> t2 = ClusterType("Polyketide")        # single type
            >>> t3 = ClusterType("Polyketide", "NRP") # multiple types

        """
        self.names = frozenset(names)

    def __repr__(self) -> str:  # noqa: D105
        return "ClusterType({})".format(", ".join(map(repr, sorted(self.names))))

    def __str__(self) -> str: # noqa: D105
        return "Unknown" if not self else ";".join(sorted(self.names))

    def __hash__(self) -> int:  # noqa: D105
        return hash(self.names)

    def __eq__(self, other: object) -> bool:  # noqa: D105
        if not isinstance(other, ClusterType):
            return NotImplemented
        return self.names == other.names

    def __bool__(self) -> bool:  # noqa: D105
        return len(self.names) != 0

    def unpack(self) -> List["ClusterType"]:
        """Unpack a composite `ClusterType` into a list of individual types.

        Example:
            >>> ty = ClusterType("Polyketide", "Saccharide")
            >>> ty.unpack()
            [ClusterType('Polyketide'), ClusterType('Saccharide')]

        """
        return [ ClusterType(x) for x in sorted(self.names) ]


class Strand(enum.IntEnum):
    """A flag to declare on which DNA strand a gene is located.
    """

    Coding = 1
    Reverse = -1

    @property
    def sign(self) -> str:
        """`str`: The strand as a single sign (``+`` or ``-``).
        """
        return "+" if self is Strand.Coding else "-"


@dataclass(frozen=True)
class Domain:
    """A conserved region within a protein.

    Attributes:
        name (`str`): The accession of the protein domain in the source HMM.
        start (`int`): The start coordinate of the domain within the protein
            sequence (first amino-acid at 1).
        end (`int`): The end coordinate of the domain within the protein
            sequence (inclusive).
        hmm (`str`): The name of the HMM library this domain belongs to
            (e.g. ``Pfam``, ``Panther``).
        i_evalue (`float`): The independent e-value reported by ``hmmsearch``
            that measures how reliable the domain annotation is.
        pvalue (`float`): The p-value reported by ``hmmsearch`` that measure
            how likely the domain score is.
        probability (`float`, optional): The probability that this domain
            is part of a gene cluster, or `None` if no prediction has been 
            made yet.
        cluster_weight (`float`, optional): The weight for this domain, 
            measuring its importance as infered from the training clusters
            by the CRF model.
        go_terms (`list` of `GOTerm`): The Gene Ontology terms
            for this particular domain.
        go_functions (`list` of `GOTerm`): The Gene Ontology term families for 
            this particular domain. Term families are extracted by taking the 
            highest superclasses (excluding the root) of each Gene Ontology 
            term in the ``molecular_function`` namespace associated with this 
            domain.
        qualifiers (`dict`): A dictionary of feature qualifiers that
            is added to the `~Bio.SeqFeature.SeqFeature` built from this
            `Domain`.

    """

    name: str
    start: int
    end: int
    hmm: str
    i_evalue: float
    pvalue: float
    probability: Optional[float] = None
    cluster_weight: Optional[float] = None
    go_terms: List[GOTerm] = field(default_factory=list)
    go_functions: List[GOTerm] = field(default_factory=list)
    qualifiers: Dict[str, List[str]] = field(default_factory=dict)

    def with_probability(self, probability: Optional[float]) -> "Domain":
        """Copy the current domain and assign it a cluster probability.
        """
        return Domain(
            self.name, self.start, self.end, self.hmm, self.i_evalue, self.pvalue,
            probability,
            self.cluster_weight,
            self.go_terms,
            self.go_functions,
            self.qualifiers.copy()
        )

    def with_cluster_weight(self, cluster_weight: Optional[float]) -> "Domain":
        """Copy the current domain and assign it a cluster weight.
        """
        return Domain(
            self.name, self.start, self.end, self.hmm, self.i_evalue, self.pvalue,
            self.probability,
            cluster_weight,
            self.go_terms,
            self.go_functions,
            self.qualifiers.copy()
        )

    def to_seq_feature(self, protein_coordinates: bool = False) -> SeqFeature:
        """Convert the domain to a single feature.

        Arguments:
            protein_coordinates (`bool`): Set to `True` for the feature
                coordinates to be given in amino-acids, or to `False` in
                nucleotides.

        """
        stride = 1 if protein_coordinates else 3
        loc = FeatureLocation((self.start-1)*stride, self.end*stride)
        qualifiers = dict(self.qualifiers)
        qualifiers.setdefault("standard_name", [self.name])
        for go_term in self.go_terms:
            qualifiers.setdefault("db_xref", []).append(go_term.accession)
        return SeqFeature(location=loc, type="misc_feature", qualifiers=qualifiers)


@dataclass(frozen=True)
class Protein:
    """A sequence of amino-acids translated from a gene.

    Attributes:
        id (`str`): The identifier of the protein.
        seq (`~Bio.Seq.Seq`): The sequence of amino-acids of this protein.
        domains (`list` of `~gecco.model.Domain`): A list of domains found
            in the protein sequence.

    """

    id: str
    seq: Seq
    domains: List[Domain] = field(default_factory=list)

    def to_seq_record(self) -> SeqRecord:
        """Convert the protein to a single record.
        """
        # FIXME: add domains
        record = SeqRecord(self.seq, id=self.id, name=self.id)

        biopython_version = tuple(map(int, Bio.__version__.split(".")))
        if biopython_version < (1, 77):
            from Bio import Alphabet

            record.seq.alphabet = Alphabet.generic_protein

        return record

    def with_seq(self, seq: Seq) -> "Protein":
        """Copy the current protein and assign it a new sequence.
        """
        return Protein(self.id, seq, copy.deepcopy(self.domains))

    def with_domains(self, domains: Iterable[Domain]) -> "Protein":
        """Copy the current protein and assign it new domains.
        """
        return Protein(self.id, self.seq, list(domains))


@dataclass(frozen=True)
class Gene:
    """A nucleotide sequence coding a protein.

    Attributes:
        source (`~Bio.SeqRecord.SeqRecord`): The DNA sequence this gene was
            found in, as a Biopython record.
        start (`int`): The index of the leftmost nucleotide of the gene within
            the source sequence, independent of the strandedness.
        end (`int`): The index of the rightmost nucleotide of the gene within
            the source sequence.
        strand (`~gecco.model.Strand`): The strand where the gene is located.
        protein (`~gecco.model.Protein`): The protein translated from this
            gene.
        qualifiers (`dict`, optional): A dictionary of feature qualifiers that
            is added to the `~Bio.SeqFeature.SeqFeature` built from this
            `Gene`.

    """

    source: SeqRecord
    start: int
    end: int
    strand: Strand
    protein: Protein
    qualifiers: Dict[str, List[str]] = field(default_factory=dict)
    _probability: Optional[float] = field(default_factory=lambda: None)

    @property
    def id(self) -> str:
        """`str`: The identifier of the gene (same as the protein identifier).
        """
        return self.protein.id

    @property
    def average_probability(self) -> Optional[float]:
        """`float`: The average of domain probabilities of being in a cluster.
        """
        if self._probability is not None:
            return self._probability
        p = [d.probability for d in self.protein.domains if d.probability is not None]
        return statistics.mean(p) if p else None

    @property
    def maximum_probability(self) -> Optional[float]:
        """`float`: The highest of domain probabilities of being in a cluster.
        """
        if self._probability is not None:
            return self._probability
        p = [d.probability for d in self.protein.domains if d.probability is not None]
        return max(p) if p else None

    # ---

    # NB(@althonos): Color scheme taken from MIBiG.
    #                This is sorted by priority!
    _FUNCTION_PALETTE = {
        # transporter: blue
        "transporter activity": (0x64, 0x95, 0xed),
        "cargo receptor activity": (0x64, 0x95, 0xed),
        "molecular carrier activity": (0x64, 0x95, 0xed),
        # regulatory: green
        "translation regulator activity": (0x2e, 0x8b, 0x56),
        "molecular function regulator activity": (0x2e, 0x8b, 0x56),
        "transcription regulator activity": (0x2e, 0x8b, 0x56),
        "regulation of molecular function": (0x2e, 0x8b, 0x56),
        "general transcription initiation factor activity": (0x2e, 0x8b, 0x56),
        # core biosynthetic: red
        "toxin activity": (0x81, 0x0e, 0x15),
        "catalytic activity": (0x81, 0x0e, 0x15),
        # additional biosynthetic: pink
        "biosynthetic activity": (0xf1, 0x6d, 0x75),
        # non-biosynthetic: olive green
        "non-biosynthetic activity": (0xbd, 0xb7, 0x6b),
        # unknown: grey
        "unknown": (0x80, 0x80, 0x80),
    }

    def to_seq_feature(self, color: bool = True) -> SeqFeature:
        """Convert the gene to a single feature.
        """
        # NB(@althonos): we use inclusive 1-based ranges in the data model
        # but Biopython expects 0-based ranges with exclusive ends
        loc = FeatureLocation(start=self.start, end=self.end+1, strand=int(self.strand))
        qualifiers = dict(self.qualifiers)
        qualifiers.setdefault("locus_tag", [self.protein.id])
        qualifiers.setdefault("translation", [str(self.protein.seq)])

        # NB(@althonos): Attempt to assign a function for the gene based on the
        #                domain content.
        functions = self.functions()
        qualifiers.setdefault("function", sorted(functions))
        if color:
            for k, hexcolor in self._FUNCTION_PALETTE.items():
                if k in functions:
                    break
            else:
                hexcolor = self._FUNCTION_PALETTE["unknown"]
            # EasyFig qualifier
            qualifiers.setdefault("colour", [" ".join(str(x) for x in hexcolor)])
            # SnapGene qualifiers
            qualifiers.setdefault("ApEinfo_fwdcolor", ["#{:02x}{:02x}{:02x}".format(*hexcolor)])
            qualifiers.setdefault("ApEinfo_revcolor", ["#{:02x}{:02x}{:02x}".format(*hexcolor)])

        return SeqFeature(location=loc, type="CDS", qualifiers=qualifiers)

    def with_protein(self, protein: "Protein") -> "Gene":
        """Copy the current gene and assign it a different protein.
        """
        return Gene(
            self.source, self.start, self.end, self.strand, protein,
            self.qualifiers.copy(),
            _probability=self._probability,
        )

    def with_source(self, source: "SeqRecord") -> "Gene":
        """Copy the current gene and assign it a different source.
        """
        return Gene(
            source, self.start, self.end, self.strand, self.protein,
            self.qualifiers.copy(),
            _probability=self._probability,
        )

    def with_probability(self, probability: float) -> "Gene":
        """Copy the current gene and assign it a different probability.
        """
        return Gene(
            self.source, self.start, self.end, self.strand,
            self.protein.with_domains([
                domain.with_probability(probability)
                for domain in self.protein.domains
            ]),
            self.qualifiers.copy(),
            _probability=probability
        )

    def functions(self) -> Set[str]:
        """Predict the function(s) of the gene from its domain annotations.
        """
        functions = {
            term.name
            for domain in self.protein.domains
            for term in domain.go_functions
        }
        if not functions:
            functions.add("unknown")
        return functions


@dataclass
class Cluster:
    """A sequence of contiguous genes.

    Attributes:
        id (`str`): The identifier of the gene cluster.
        genes (`list` of `~gecco.model.Gene`): A list of the genes belonging
            to this gene cluster.
        types (`gecco.model.ClusterType`): The putative types of this gene 
            cluster, according to similarity in domain composition with 
            curated clusters.
        types_probabilities (`list` of `float`): The probability with which
            each cluster type was identified (same dimension as the ``types``
            attribute).

    """

    id: str
    genes: List[Gene]
    type: Optional[ClusterType]
    type_probabilities: Dict[str, float]

    def __init__(
        self,
        id: str,
        genes: Optional[List[Gene]] = None,
        type: Optional[ClusterType] = None,
        type_probabilities: Optional[Dict[str, float]] = None,
    ):  # noqa: D107
        self.id = id
        self.genes = genes or list()
        self.type = type
        self.type_probabilities = type_probabilities or dict()

    @property
    def source(self) -> SeqRecord:  # type: ignore
        """`~Bio.SeqRecord.SeqRecord`: The sequence this cluster was found in.
        """
        return self.genes[0].source

    @property
    def start(self) -> int:
        """`int`: The start of this cluster in the source sequence.
        """
        return min(gene.start for gene in self.genes)

    @property
    def end(self) -> int:
        """`int`: The end of this cluster in the source sequence.
        """
        return max(gene.end for gene in self.genes)

    @property
    def average_probability(self) -> Optional[float]:
        """`float`: The average of proteins probability of being biosynthetic.
        """
        p = [g.average_probability for g in self.genes if g.average_probability is not None]
        return  statistics.mean(p) if p else None

    @property
    def maximum_probability(self) -> Optional[float]:
        """`float`: The highest of proteins probability of being biosynthetic.
        """
        p = [g.maximum_probability for g in self.genes if g.maximum_probability is not None]
        return max(p) if p else None

    # ---

    def domain_composition(
        self,
        all_possible: Optional[Sequence[str]] = None,
        normalize: bool = True,
        minlog_weights: bool = False,
        pvalue: bool = True,
    ) -> "NDArray[numpy.double]":
        """Compute weighted domain composition with respect to ``all_possible``.

        Arguments:
            all_possible (sequence of `str`, optional): A sequence containing
                all domain names to consider when computing domain composition
                for the cluster. If `None` given, then only domains within the
                cluster are taken into account.
            normalize (`bool`): Normalize the composition vector so that it
                sums to 1.
            minlog_weights (`bool`): Compute weight for each domain as
                :math:`-log_10(v)` (where :math:`v` is either the ``pvalue``
                or the ``i_evalue``, depending on the value of ``normalize``).
                Use :math:`1 - v` otherwise.
            pvalue (`bool`): Compute composition weights using the ``pvalue``
                of each domain, instead of the ``i_evalue``.

        Returns:
            `~numpy.ndarray`: A numerical array containing the relative domain
            composition of the gene cluster.

        """
        domains = [d for gene in self.genes for d in gene.protein.domains]
        names = numpy.array([domain.name for domain in domains])
        field = operator.attrgetter("pvalue" if pvalue else "i_evalue")
        if minlog_weights:
            weights = numpy.array([- math.log10(field(domain)) for domain in domains])
        else:
            weights = numpy.array([1 - field(domain) for domain in domains])

        unique_names = set(names)
        if all_possible is None:
            all_possible = numpy.unique(names)
        composition = numpy.zeros(len(all_possible))
        for i, dom in enumerate(all_possible):
            if dom in unique_names:
                composition[i] = numpy.sum(weights[names == dom])
        if normalize:
            return composition / (composition.sum() or 1)
        return composition

    # ---

    def to_seq_record(self) -> SeqRecord:
        """Convert the cluster to a single record.

        Annotations of the source sequence are kept intact if they don't
        overlap with the cluster boundaries. Component genes are added on the
        record as *CDS* features. Annotated protein domains are added as
        *misc_feature*.

        """
        # store time of record creation
        now = datetime.datetime.now()

        # NB(@althonos): we use inclusive 1-based ranges in the data model
        # but slicing expects 0-based ranges with exclusive ends
        cluster = self.source[self.start - 1 : self.end]
        cluster.id = cluster.name = self.id

        # copy sequence annotations
        cluster.annotations = self.source.annotations.copy()
        cluster.annotations["topology"] = "linear"
        cluster.annotations["molecule_type"] = "DNA"
        with patch_locale("C"):
            cluster.annotations['date'] = now.strftime("%d-%b-%Y").upper()

        biopython_version = tuple(map(int, Bio.__version__.split(".")))
        if biopython_version < (1, 77):
            from Bio import Alphabet

            cluster.seq.alphabet = Alphabet.generic_dna

        # add GECCO preprint as a reference
        ref = Reference()
        ref.title = "Accurate de novo identification of biosynthetic gene clusters with GECCO"
        ref.journal = "bioRxiv (2021.05.03.442509)"
        ref.comment = "doi:10.1101/2021.05.03.442509"
        ref.authors = ", ".join([
            "Laura M Carroll",
            "Martin Larralde",
            "Jonas Simon Fleck",
            "Ruby Ponnudurai",
            "Alessio Milanese",
            "Elisa Cappio Barazzone",
            "Georg Zeller"
        ])
        cluster.annotations.setdefault("references", []).append(ref)

        # add GECCO-specific annotations as a structured comment
        if self.type is not None:
            cluster_type = {
                "cluster_type": ";".join(sorted(self.type.names)) or "Unknown"
            }
            probabilities = {
                f"{key.lower()}_probability":f"{value:.3f}"
                for key, value in self.type_probabilities.items()
            }
        else:
            cluster_type = probabilities = {}

        structured_comment = cluster.annotations.setdefault("structured_comment", OrderedDict())
        structured_comment['GECCO-Data'] = {
            "version": f"GECCO v{__version__}",
            "creation_date": now.isoformat(),
            **cluster_type,
            **probabilities,
        }

        # add proteins as CDS features
        for gene in self.genes:
            # write gene as a /cds GenBank record
            cds = gene.to_seq_feature()
            cds.location += -self.start
            cluster.features.append(cds)
            # write domains as /misc_feature annotations
            for domain in gene.protein.domains:
                misc = domain.to_seq_feature(protein_coordinates=False)
                if gene.strand == Strand.Coding:
                    misc.location = FeatureLocation(
                        cds.location.start + misc.location.start,
                        cds.location.start + misc.location.end
                    )
                else:
                    misc.location = FeatureLocation(
                        cds.location.end - misc.location.end,
                        cds.location.end - misc.location.start
                    )
                cluster.features.append(misc)

        # return the complete gene cluster record
        return cluster


class _UnknownSeq(Seq):
    """An unknown sequence that uses limited memory.

    Used by `FeatureTable.to_genes` to fake the `Gene.source.seq` attribute.
    """

    def __init__(self) -> None:
        super().__init__(data="")

    @typing.overload
    def __getitem__(self, index: int) -> str:
        pass

    @typing.overload
    def __getitem__(self, index: slice) -> Seq:
        pass

    def __getitem__(self, index: Union[slice, int]) -> Union[str, Seq]:
        if isinstance(index, slice):
            return Seq("N" * ((index.stop - index.start) // (index.step or 1)) )
        return "N"


class FeatureTable(Table):
    """A table storing condensed domain annotations from different genes.
    """

    @classmethod
    def _get_columns(cls) -> List["Table.Column"]:
        return [
            Table.Column("sequence_id", polars.Utf8),
            Table.Column("protein_id", polars.Utf8),
            Table.Column("start", polars.Int64),
            Table.Column("end", polars.Int64),
            Table.Column("strand", polars.Utf8),
            Table.Column("domain", polars.Utf8),
            Table.Column("hmm", polars.Utf8),
            Table.Column("i_evalue", polars.Float64),
            Table.Column("pvalue", polars.Float64),
            Table.Column("domain_start", polars.Int64),
            Table.Column("domain_end", polars.Int64),
            Table.Column("cluster_probability", polars.Float64, default=math.nan),
        ]

    @classmethod
    def from_genes(cls, genes: Iterable[Gene]) -> "FeatureTable":
        """Create a new feature table from an iterable of genes.
        """
        columns = cls._get_columns()
        data = { column.name: [] for column in columns }
        for gene in genes:
            for domain in gene.protein.domains:
                data["sequence_id"].append(gene.source.id)
                data["protein_id"].append(gene.protein.id)
                data["start"].append(gene.start)
                data["end"].append(gene.end)
                data["strand"].append(gene.strand.sign)
                data["domain"].append(domain.name)
                data["hmm"].append(domain.hmm)
                data["i_evalue"].append(domain.i_evalue)
                data["pvalue"].append(domain.pvalue)
                data["domain_start"].append(domain.start)
                data["domain_end"].append(domain.end)
                if domain.probability is not None:
                    data["cluster_probability"].append(domain.probability)
        for name in list(data):
            if not data[name]:
                del data[name]
        return cls(polars.DataFrame(data))

    def to_genes(self) -> Iterable[Gene]:
        """Convert a feature table to actual genes.

        Since the source sequence cannot be known, a *dummy* sequence is
        built for each gene of size ``gene.end``, so that each gene can still
        be converted to a `~Bio.SeqRecord.SeqRecord` if needed.

        """
        # group rows by protein/gene ID
        protein_indices = collections.defaultdict(list)
        for i, protein_id in enumerate(self.protein_id):
            protein_indices[protein_id].append(i)
        # yield genes in order
        for protein_id in sorted(protein_indices):
            indices = protein_indices[protein_id]
            assert all(self.sequence_id[i] == self.sequence_id[indices[0]] for i in indices)
            assert all(self.protein_id[i] == self.protein_id[indices[0]] for i in indices)
            assert all(self.start[i] == self.start[indices[0]] for i in indices)
            assert all(self.end[i] == self.end[indices[0]] for i in indices)
            source = SeqRecord(id=self.sequence_id[indices[0]], seq=_UnknownSeq())
            strand = Strand.Coding if self.strand[indices[0]] == "+" else Strand.Reverse
            protein = Protein(self.protein_id[indices[0]], seq=_UnknownSeq)
            gene = Gene(source, self.start[indices[0]], self.end[indices[0]], strand, protein)
            for i in indices:
                domain = Domain(
                    self.domain[i], 
                    self.domain_start[i], 
                    self.domain_end[i],
                    self.hmm[i], 
                    self.i_evalue[i], 
                    self.pvalue[i], 
                    self.cluster_probability[i]
                )
                gene.protein.domains.append(domain)
            yield gene


class ClusterTable(Table):
    """A table storing condensed information from several clusters.
    """

    @classmethod
    def _get_columns(cls) -> List["Table.Column"]:
        return [
            Table.Column("sequence_id", polars.Utf8),
            Table.Column("cluster_id", polars.Utf8),
            Table.Column("start", polars.Int64),
            Table.Column("end", polars.Int64),
            Table.Column("average_p", polars.Float64, default=math.nan),
            Table.Column("max_p", polars.Float64, default=math.nan),
            Table.Column("type", polars.Utf8, default="Unknown"),
            # + possible type columns that are handled in `from_clusters`
            Table.Column("proteins", polars.Utf8, default=""),
            Table.Column("domains", polars.Utf8, default=""),
        ]

    @classmethod
    def from_clusters(cls, clusters: Iterable[Cluster]) -> "ClusterTable":
        """Create a new cluster table from an iterable of clusters.
        """
        data = collections.defaultdict(list)
        for cluster in clusters:
            data["sequence_id"].append(cluster.source.id)
            data["cluster_id"].append(cluster.id)
            data["start"].append(cluster.start)
            data["end"].append(cluster.end)
            if cluster.average_probability is not None:
                data["average_p"].append(cluster.average_probability)
            if cluster.maximum_probability is not None:
                data["max_p"].append(cluster.maximum_probability)
            if cluster.type is not None:
                data["type"].append(str(cluster.type))
                for type_name in sorted(cluster.type_probabilities, key=str.casefold):
                    data[f"{type_name.lower()}_probability"].append(cluster.type_probabilities[type_name])
            data["proteins"].append(";".join(
                sorted(gene.protein.id for gene in cluster.genes)
            ))
            data["domains"].append(";".join(sorted( 
                domain.name 
                for gene in cluster.genes 
                for domain in gene.protein.domains 
            )))
            # TODO: member proteins
            # TODO: member domains
        return cls(polars.DataFrame(data))

    def dump(self, fh: Union[BinaryIO, str, os.PathLike]) -> None:
        # patch `Table.dump` so that all columns are always written
        data = self.data
        for column_name in data.columns:
            if data[column_name].dtype in (polars.Float64, polars.Float64):
                data = data.with_columns(polars.col(column_name).fill_nan(None))
        if _POLARS_VERSION < (0, 16, 14):
            data.write_csv(fh, sep="\t")
        else:
            data.write_csv(fh, separator="\t")


class GeneTable(Table):
    """A table storing gene coordinates and optional cluster probabilities.
    """

    @classmethod
    def _get_columns(cls) -> List["Table.Column"]:
        return [
            Table.Column("sequence_id", polars.Utf8),
            Table.Column("protein_id", polars.Utf8),
            Table.Column("start", polars.Int64),
            Table.Column("end", polars.Int64),
            Table.Column("strand", polars.Utf8),
            Table.Column("average_p", polars.Float64, default=math.nan),
            Table.Column("max_p", polars.Float64, default=math.nan),
        ]

    @classmethod
    def from_genes(cls, genes: Iterable[Gene]) -> "GeneTable":
        """Create a new gene table from an iterable of genes.
        """
        columns = cls._get_columns()
        data = { column.name: [] for column in columns }
        for gene in genes:
            data["sequence_id"].append(gene.source.id)
            data["protein_id"].append(gene.protein.id)
            data["start"].append(gene.start)
            data["end"].append(gene.end)
            data["strand"].append(gene.strand.sign)
            if gene.average_probability is not None:
                data["average_p"].append(gene.average_probability)
            else:
                data["average_p"].append(math.nan)
            if gene.maximum_probability is not None:
                data["max_p"].append(gene.maximum_probability)
            else:
                data["max_p"].append(math.nan)
        return cls(polars.DataFrame(data))

    def to_genes(self) -> Iterable[Gene]:
        """Convert a gene table to actual genes.

        Since the source sequence cannot be known, a *dummy* sequence is
        built for each gene of size ``gene.end``, so that each gene can still
        be converted to a `~Bio.SeqRecord.SeqRecord` if needed.

        """
        # check if a probability column is available 
        has_probas = "average_p" in self.data.columns
        # yield genes in order
        for i, protein_id in enumerate(self.protein_id):
            source = SeqRecord(id=self.sequence_id[i], seq=_UnknownSeq())
            strand = Strand.Coding if self.strand[i] == "+" else Strand.Reverse
            start = self.start[i]
            end = self.end[i]
            seq = Seq("X" * ((end - start) // 3))
            protein = Protein(self.protein_id[i], seq=seq)
            probability = self.average_p[i] if has_probas else None
            gene = Gene(source, start, end, strand, protein, _probability=probability)
            yield gene



"""Simple data classes to expose embedded InterPro data.
"""

import gzip
import json
from dataclasses import dataclass, field, fields
from typing import Dict, List, Optional

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files  # type: ignore


__all__ = ["InterProEntry", "InterPro", "GeneOntologyTerm"]




@dataclass
class GOTerm:
    """A single term from the Gene Ontology.
    """
    accession: str
    name: str
    namespace: str

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
                self.accession == other.accession
            and self.name == other.name
            and self.namespace == other.namespace
        )

    def __hash__(self) -> int:
        return hash((type(self), self.accession, self.name, self.namespace))


@dataclass
class InterProEntry:
    """A single entry in the InterPro database.
    """

    accession: str
    members: List[str]
    name: str
    databases: List[str]
    type: str
    go_terms: List[GOTerm]
    go_functions: List[GOTerm]


@dataclass
class InterPro:
    """A subset of the InterPro database exposing domain metadata.
    """

    entries: List[InterProEntry]

    def __init__(self, entries: List[InterProEntry]):
        self.entries = entries
        self.by_accession = { member:entry for entry in entries for member in entry.members }

    @classmethod
    def load(cls) -> "InterPro":
        with files(__name__).joinpath("interpro.json").open() as f:
            data = json.load(f)
            entries = []
            for raw_entry in data:
                # get terms corresponding to domain
                go_terms = [
                    GOTerm(**t) for t in raw_entry.pop("go_terms")
                ]
                go_functions = [
                    GOTerm(**t, namespace="molecular_function") 
                    for t in raw_entry.pop("go_functions")
                ]
                entries.append(
                    InterProEntry(**raw_entry, go_terms=go_terms, go_functions=go_functions)
                )
        return cls(entries)



"""Implementation of the ``gecco annotate`` subcommand.
"""

import contextlib
import errno
import glob
import itertools
import io
import logging
import multiprocessing
import operator
import os
import pickle
import tempfile
import typing
import signal
from typing import Any, BinaryIO, Container, Dict, Iterable, Union, Optional, List, TextIO, Mapping

from ._base import Command, CommandExit, InvalidArgument
from ._mixins import SequenceLoaderMixin, OutputWriterMixin, AnnotatorMixin
from .._utils import (
    guess_sequences_format,
    in_context,
    patch_showwarnings,
    ProgressReader,
)

if typing.TYPE_CHECKING:
    from Bio.SeqRecord import SeqRecord
    from ...hmmer import HMM
    from ...model import Gene
    from ...orf import ORFFinder


class Annotate(SequenceLoaderMixin, OutputWriterMixin, AnnotatorMixin):  # noqa: D101

    summary = "annotate protein features of one or several contigs."

    @classmethod
    def doc(cls, fast: bool = False) -> str:  # noqa: D102
        return f"""
        gecco annotate - {cls.summary}

        Usage:
            gecco annotate --genome <file> [--hmm <hmm>]... [--hmm-x <hmm>]... [options]

        Arguments:
            -g <file>, --genome <file>    a genomic file containing one or more
                                          sequences to use as input. Must be in
                                          one of the sequences format supported
                                          by Biopython.

        Parameters:
            -f <fmt>, --format <fmt>      the format of the input file, as a
                                          Biopython format string. GECCO is able
                                          to recognize FASTA and GenBank files
                                          automatically if this is not given.
            -j <jobs>, --jobs <jobs>      the number of CPUs to use for
                                          multithreading. Use 0 to use all of
                                          the available CPUs. [default: 0]

        Parameters - Output:
            -o <out>, --output-dir <out>  the directory in which to write the
                                          output files. [default: .]
            --force-tsv                   always write TSV output files even
                                          when they are empty (e.g. because
                                          no genes or no clusters were found).


        Parameters - Gene Calling:
            -M, --mask                    Enable unknown region masking to
                                          prevent genes from stretching across
                                          unknown nucleotides.
            --cds-feature <cds_feature>   Extract genes from annotated records
                                          using a feature rather than calling
                                          genes from scratch.
            --locus-tag <locus_tag>       The name of the feature qualifier
                                          to use for naming extracted genes
                                          when using the ``--cds-feature``
                                          flag. [default: locus_tag]

        Parameters - Domain Annotation:
            --hmm <hmm>                   the path to one or more alternative
                                          HMM file to use (in HMMER format).
            -e <e>, --e-filter <e>        the e-value cutoff for protein domains
                                          to be included. This is not stable
                                          across versions, so consider using
                                          a p-value filter instead.
            -p <p>, --p-filter <p>        the p-value cutoff for protein domains
                                          to be included. [default: 1e-9]
            --bit-cutoffs <name>          use bitscore cutoffs (one of *noise*,
                                          *gathering*, or *trusted*) to filter
                                          domain annotations.
            --disentangle                 disentangle overlapping domains in 
                                          each gene by keeping only the domains
                                          with the lowest E-value over a given
                                          position.

        """

    def _check(self) -> None:
        _BITSCORE_CUTOFFS = {"gathering", "noise", "trusted"}
        super()._check()
        try:
            self.e_filter = self._check_flag(
                "--e-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.p_filter = self._check_flag(
                "--p-filter",
                float,
                lambda x: x > 0,
                hint="real number above 0",
                optional=True,
            )
            self.jobs = self._check_flag("--jobs", int, lambda x: x >= 0, hint="positive or null integer")
            self.format: Optional[str] = self._check_flag("--format", optional=True)
            self.genome: str = self._check_flag("--genome")
            self.hmm: Optional[List[str]] = self._check_flag("--hmm", optional=True)
            self.output_dir: str = self._check_flag("--output-dir")
            self.mask = self._check_flag("--mask", bool)
            self.force_tsv = self._check_flag("--force-tsv", bool)
            self.cds_feature: Optional[str] = self._check_flag("--cds-feature", optional=True)
            self.locus_tag: str = self._check_flag("--locus-tag")
            self.disentangle = self._check_flag("--disentangle", bool)
            self.bit_cutoffs: str = self._check_flag(
                "--bit-cutoffs",
                str,
                _BITSCORE_CUTOFFS.__contains__,
                optional=True,
                hint="one of {}".format(", ".join(sorted(_BITSCORE_CUTOFFS)))
            )
        except InvalidArgument:
            raise CommandExit(1)

    # ---

    _OUTPUT_FILES = ["features.tsv", "genes.tsv"]

    def _extract_genes(self, sequences: List["SeqRecord"]) -> List["Gene"]:
        from ...orf import PyrodigalFinder, CDSFinder

        self.info("Extracting", "genes from input sequences", level=1)
        if self.cds_feature is None:
            self.info("Using", "Pyrodigal in metagenomic mode", level=2)
            orf_finder: ORFFinder = PyrodigalFinder(metagenome=True, mask=self.mask, cpus=self.jobs)
        else:
            self.info("Using", f"record features named {self.cds_feature!r}", level=2)
            orf_finder = CDSFinder(feature=self.cds_feature, locus_tag=self.locus_tag)

        unit = "contigs" if len(sequences) > 1 else "contig"
        task = self.progress.add_task(description="Finding ORFs", total=len(sequences), unit=unit, precision="")

        def callback(record: "SeqRecord", found: int) -> None:
            self.success("Found", found, "genes in record", repr(record.id), level=2)
            self.progress.update(task, advance=1)

        return list(orf_finder.find_genes(sequences, progress=callback))

    # ---

    def execute(self, ctx: contextlib.ExitStack) -> int:  # noqa: D102
        try:
            # check the CLI arguments were fine and enter context
            self._check()
            ctx.enter_context(self.progress)
            ctx.enter_context(patch_showwarnings(self._showwarnings))  # type: ignore
            # attempt to create the output directory, checking it doesn't
            # already contain output files (or raise a warning)
            base, _ = os.path.splitext(os.path.basename(self.genome))
            outputs = [f"{base}.features.tsv", f"{base}.genes.tsv"]
            self._make_output_directory(outputs)
            # load sequences and extract genes
            sequences = list(self._load_sequences())
            genes = self._extract_genes(sequences)
            self._write_genes_table(genes)
            if genes:
                self.success("Found", "a total of", len(genes), "genes", level=1)
            else:
                if self.force_tsv:
                    self._write_feature_table([])
                self.warn("No genes were found")
                return 0
            # annotate domains and write results
            genes = self._annotate_domains(genes)
            self._write_feature_table(genes)
            ndoms = sum(1 for gene in genes for domain in gene.protein.domains)
            # report number of proteins found
            if ndoms:
                self.success("Found", ndoms, "protein domains", level=0)
            else:
                self.warn("No protein domains were found")
        except CommandExit as cexit:
            return cexit.code
        except KeyboardInterrupt:
            self.error("Interrupted")
            return -signal.SIGINT
        except Exception as err:
            self.progress.stop()
            raise
        else:
            return 0



"""Compatibility wrapper for HMMER binaries and output.
"""
import abc
import atexit
import collections
import configparser
import contextlib
import csv
import errno
import glob
import gzip
import itertools
import os
import re
import subprocess
import tempfile
import typing
from typing import Any, BinaryIO, Callable, Container, Dict, Optional, Iterable, Iterator, List, Mapping, Type, Sequence

import pyhmmer
from Bio import SeqIO
from pyhmmer.hmmer import hmmsearch

from .._meta import requires, UniversalContainer, zopen
from ..model import Gene, Domain
from ..interpro import InterPro

try:
    from importlib.resources import files, as_file
except ImportError:
    from importlib_resources import files, as_file  # type: ignore

__all__ = ["DomainAnnotator", "HMM", "PyHMMER", "embedded_hmms"]


class HMM(typing.NamedTuple):
    """A Hidden Markov Model library to use with `~gecco.hmmer.HMMER`.
    """

    id: str
    version: str
    url: str
    path: str
    size: Optional[int] = None
    relabel_with: Optional[str] = None
    md5: Optional[str] = None

    def relabel(self, domain: str) -> str:
        """Rename a domain obtained by this HMM to the right accession.

        This method can be used with HMM libraries that have separate HMMs
        for the same domain, such as Pfam.
        """
        if self.relabel_with is None:
            return domain
        before, after = re.match("^s/(.*)/(.*)/$", self.relabel_with).groups()  # type: ignore
        regex = re.compile(before)
        return regex.sub(after, domain)


class DomainAnnotator(metaclass=abc.ABCMeta):
    """An abstract class for annotating genes with protein domains.
    """

    def __init__(self, hmm: HMM, cpus: Optional[int] = None, whitelist: Optional[Container[str]] = None) -> None:
        """Prepare a new HMMER annotation handler with the given ``hmms``.

        Arguments:
            hmm (str): The path to the file containing the HMMs.
            cpus (int, optional): The number of CPUs to allocate for the
                ``hmmsearch`` command. Give ``None`` to use the default.
            whitelist (container of str): If given, a container containing
                the accessions of the individual HMMs to annotate with. If
                `None` is given, annotate with the entire file.

        """
        super().__init__()
        self.hmm = hmm
        self.cpus = cpus
        self.whitelist = UniversalContainer() if whitelist is None else whitelist

    @abc.abstractmethod
    def run(self, genes: Iterable[Gene]) -> List[Gene]:
        """Run annotation on proteins of ``genes`` and update their domains.

        Arguments:
            genes (iterable of `~gecco.model.Gene`): An iterable that yield
                genes to annotate with ``self.hmm``.

        """
        return NotImplemented


class PyHMMER(DomainAnnotator):
    """A domain annotator that uses `pyhmmer`.
    """

    def run(
        self,
        genes: Iterable[Gene],
        progress: Optional[Callable[[pyhmmer.plan7.HMM, int], None]] = None,
        bit_cutoffs: Optional[str] = None,
    ) -> List[Gene]:
        # collect genes and keep them in original order
        gene_index = list(genes)

        # convert proteins to Easel sequences, namind them after
        # their location in the original input to ignore any duplicate
        # protein identifiers
        esl_abc = pyhmmer.easel.Alphabet.amino()
        esl_sqs = [
            pyhmmer.easel.TextSequence(
                name=str(i).encode(),
                sequence=str(gene.protein.seq)
            ).digitize(esl_abc)
            for i, gene in enumerate(gene_index)
        ]

        with contextlib.ExitStack() as ctx:
            # decompress the input if needed
            file: BinaryIO = ctx.enter_context(zopen(self.hmm.path))
            # Only retain the HMMs which are in the whitelist
            hmm_file = ctx.enter_context(pyhmmer.plan7.HMMFile(file))
            profiles = (
                hmm
                for hmm in hmm_file
                if hmm.accession is None
                or self.hmm.relabel(hmm.accession.decode()) in self.whitelist
            )
            # Run search pipeline using the filtered HMMs
            cpus = 0 if self.cpus is None else self.cpus
            hmms_hits = hmmsearch(
                profiles,
                esl_sqs,
                cpus=cpus,
                callback=progress, # type: ignore
                Z=self.hmm.size,  # type: ignore
                domZ=self.hmm.size, # type: ignore
                bit_cutoffs=bit_cutoffs,  # type: ignore
            )

            # Load InterPro metadata for the annotation
            interpro = InterPro.load()

            # Transcribe HMMER hits to GECCO model
            for hits in hmms_hits:
                for hit in hits.reported:
                    target_index = int(hit.name)
                    for domain in hit.domains.reported:
                        # extract name and get InterPro metadata about hit
                        raw_acc = domain.alignment.hmm_accession or domain.alignment.hmm_name
                        accession = self.hmm.relabel(raw_acc.decode('utf-8'))
                        entry = interpro.by_accession.get(accession)

                        # extract coordinates
                        start = domain.alignment.target_from
                        end = domain.alignment.target_to

                        # extract qualifiers and GO terms
                        qualifiers: Dict[str, List[str]] = {
                            "inference": ["protein motif"],
                            "db_xref": ["{}:{}".format(self.hmm.id.upper(), accession)],
                            "note": [
                                "e-value: {}".format(domain.i_evalue),
                                "p-value: {}".format(domain.pvalue),
                            ],
                        }
                        if entry is not None:
                            qualifiers["function"] = [entry.name]
                            qualifiers["db_xref"].append("InterPro:{}".format(entry.accession))
                            go_terms = entry.go_terms
                            go_functions = entry.go_functions
                        else:
                            go_terms = []
                            go_functions = []

                        # add the domain to the protein domains of the right gene
                        assert domain.env_from < domain.env_to
                        assert domain.i_evalue >= 0
                        assert domain.pvalue >= 0
                        gene_index[target_index].protein.domains.append(
                            Domain(
                                accession,
                                start,
                                end,
                                self.hmm.id,
                                domain.i_evalue,
                                domain.pvalue,
                                go_terms=go_terms,
                                go_functions=go_functions,
                                qualifiers=qualifiers,
                            )
                        )

        # return the updated list of genes that was given in argument
        return gene_index


def embedded_hmms() -> Iterator[HMM]:
    """Iterate over the embedded HMMs that are shipped with GECCO.
    """
    for filename in files(__name__).glob("*.ini"):

        ini_ctx = as_file(filename)
        ini_path = ini_ctx.__enter__()
        atexit.register(ini_ctx.__exit__, None, None, None)

        cfg = configparser.ConfigParser()
        cfg.read(ini_path)
        args: Dict[str, Any] = dict(cfg.items("hmm"))
        size = int(args.pop("size", 0))

        hmm_ctx = as_file(filename.with_suffix(".h3m"))
        hmm_path = hmm_ctx.__enter__()
        atexit.register(hmm_ctx.__exit__, None, None, None)

        yield HMM(path=os.fspath(hmm_path), size=size, **args)
